"""
 Configuration parser module.

 Programmed by: Dante Signal31

 email: dante.signal31@gmail.com
"""

import http.client as http
import os
import pickle
import urllib.parse as urlparse


def get_real_path(CONFIG_FILE):
    this_module_path = os.path.realpath(__file__)
    this_module_folder = os.path.dirname(this_module_path)
    parent_folder, _ = os.path.split(this_module_folder)
    config_file_path = os.path.join(parent_folder, CONFIG_FILE)
    return config_file_path

CONFIG_FILE = "etc/geolocate.conf"
CONFIG_FILE_PATH = get_real_path(CONFIG_FILE)
DEFAULT_USER_ID = ""
DEFAULT_LICENSE_KEY = ""
# TODO: For production I have to uncomment real url.
# Only for tests I have to comment real download url. MaxMind has a rate limit
# per day. If you exceed that limit you are forbidden for 24 hours to download
# their database.
DEFAULT_DATABASE_DOWNLOAD_URL = "http://geolite.maxmind.com/download/geoip/database/GeoLite2-City.mmdb.gz"
# TODO: For production remove next fake url, it's only for tests.
# DEFAULT_DATABASE_DOWNLOAD_URL = "http://localhost:2014/GeoLite2-City.mmdb.gz"
# GeoLite2 databases are updated on the first Tuesday of each month, so 35 days
# of update interval should be fine.
DEFAULT_UPDATE_INTERVAL = 35
DEFAULT_LOCAL_DATABASE_FOLDER = get_real_path("local_database/")
DEFAULT_LOCAL_DATABASE_NAME = "GeoLite2-City.mmdb"
# Remember add new locators here or locate won't use them.
DEFAULT_LOCATORS_PREFERENCE = ["geoip2_webservice", "geoip2_local"]


class Configuration(object):
    # I've discovered Maxmind website blocks clients who exceeds a connection
    # threshold. If we make a connection each time we run geolocate, in order
    # to check that configured URL is OK, we can end in Maxmind blacklist. So,
    # we have to minimize connections. Check only when configuration is updated
    # is a way, but then we have to control how users update config. Best way
    # is limit users to change configuration only through executable's
    # parameters. If we let them change manually configuration file we should
    # check it each time we run the program. We'd better close configuration
    # file through serialization and check URL only when user makes program
    # change configuration.
    """ Class to encapsulate configuration needed to connect to Geolite2
    webservices or downloaded local database. This class also validates
    parameters read from config files to overcome user typos.
    """
    def __init__(self, user_id=DEFAULT_USER_ID,
                 license_key=DEFAULT_LICENSE_KEY,
                 download_url=DEFAULT_DATABASE_DOWNLOAD_URL,
                 update_interval=DEFAULT_UPDATE_INTERVAL,
                 local_database_folder=DEFAULT_LOCAL_DATABASE_FOLDER,
                 local_database_name=DEFAULT_LOCAL_DATABASE_NAME,
                 locators_preference=DEFAULT_LOCATORS_PREFERENCE):
        self._webservice = {"user_id": user_id,
                            "license_key": license_key}
        self._local_database = {"download_url": download_url,
                                "update_interval": update_interval,
                                "local_database_folder": local_database_folder,
                                "local_database_name": local_database_name}
        self._locators_preference = locators_preference

    @property
    def user_id(self):
        return self._webservice["user_id"]

    @user_id.setter
    def user_id(self, user_id):
        _validate_value("user_id", user_id)
        self._webservice["user_id"] = user_id

    @property
    def license_key(self):
        return self._webservice["license_key"]

    @license_key.setter
    def license_key(self, license_key):
        _validate_value("license_key", license_key)
        self._webservice["license_key"] = license_key

    @property
    def download_url(self):
        return self._local_database["download_url"]

    @download_url.setter
    def download_url(self, url):
        _validate_url("download_url", url)
        self._local_database["download_url"] = url

    @property
    def update_interval(self):
        return self._local_database["update_interval"]

    @update_interval.setter
    def update_interval(self, update_interval_in_days):
        interval_integer = _validate_integer("update_interval",
                                             update_interval_in_days)
        self._local_database["update_interval"] = interval_integer

    @property
    def local_database_folder(self):
        return self._local_database["local_database_folder"]

    @local_database_folder.setter
    def local_database_folder(self, folder_path):
        database_folder_path = _get_folder_path(folder_path)
        _validate_folder("local_database_folder", database_folder_path)
        self._local_database["local_database_folder"] = database_folder_path

    @property
    def local_database_name(self):
        return self._local_database["local_database_name"]

    @local_database_name.setter
    def local_database_name(self, database_name):
        # At first sight every database name should be OK, but I'll leave this
        # as a property in case I have an idea about a possible check.
        self._local_database["local_database_name"] = database_name

    @property
    def local_database_path(self):
        path = os.path.join(self.local_database_folder,
                            self.local_database_name)
        return path

    @property
    def locators_preference(self):
        """
        :return: Enabled locators for this GeoIPDatabase ordered by preference.
        :rtype: list
        """
        return self._locators_preference

    @locators_preference.setter
    def locators_preference(self, new_locator_list):
        if _unknown_locators(new_locator_list):
            unknown_locators = _get_unknown_locators(new_locator_list)
            raise UnknownLocators(unknown_locators)
        else:
            self._locators_preference = new_locator_list

    def __eq__(self, other):
        for _property, value in vars(self).items():
            if getattr(other, _property) != value:
                return False
        return True

    def reset_locators_preference(self):
        """ Reset locators preference to default order.

        :return: None
        """
        self._locators_preference = DEFAULT_LOCATORS_PREFERENCE

    @property
    def disabled_locators(self):
        """ Locators registered as default one but not enabled in this
        GeoIPDatabase.

        :return: Disabled locators in this GeoIPDatabase.
        :rtype: set
        """
        default_locators_set = set(DEFAULT_LOCATORS_PREFERENCE)
        enabled_locators_set = set(self.locators_preference)
        disabled_locators_set = default_locators_set - enabled_locators_set
        return disabled_locators_set


def _validate_value(parameter, value):
    # TODO: Add more checks to detect invalid values when you know MaxMind's
    # conditions for user ids.
    if _text_has_spaces(value) or value == "":
        raise ParameterNotValid(value, parameter,
                                " ". join([parameter, "cannot have spaces."]))


def _validate_url(parameter, url):
    """
    Check if a URL exists without downloading the whole file.
    We only check the URL header.

    :param parameter: Attribute that is being validated.
    :type parameter: str
    :param url: HTTP url to check its existence.
    :type url: str
    :return: None
    :raise: ParameterNotValid
    """
    # see also http://stackoverflow.com/questions/2924422
    good_codes = [http.OK, http.FOUND, http.MOVED_PERMANENTLY]
    try:
        if _get_server_status_code(url) in good_codes:
            return  # Validation succeeded.
        else:
            raise Exception  # Let outer except raise one only exception.
    except:
        raise ParameterNotValid(url, parameter, "Cannot connect to given "
                                                "URL.")


def _validate_folder(parameter, path):
    """
    :param parameter: Attribute that is being validated.
    :type parameter: str
    :param path: Path to folder being checked.
    :type path: str
    :return: None
    :raise: ParameterNotValid
    """
    if not os.path.exists(path):
        raise ParameterNotValid(path, parameter, "Folder does not exists.")


def _get_server_status_code(url):
    """
    Download just the header of a URL and return the server's status code.

    :param url: HTTP url to check its existence.
    :type url: str
    :return: One of the connection status from http.client.
    :rtype: int
    :raise: Any of the exceptions from http.client built-in module.
    """
    # http://stackoverflow.com/questions/1140661
    host, path = urlparse.urlparse(url)[1:3]    # elems [1] and [2]
    conn = http.HTTPConnection(host)
    conn.request('HEAD', path)
    return conn.getresponse().status


def _validate_integer(parameter, value):
    """
    :param parameter: Attribute that is being validated.
    :type parameter: str
    :param value: Value integer o string.
    :type value: int or str
    :return: Value converted to an integer.
    :rtype: int
    """
    try:
        integer_value = int(value)
        if integer_value <= 0:
            raise ValueError
    except ValueError:
        raise ParameterNotValid(value, parameter, "Cannot convert to int.")
    return integer_value


def _text_has_spaces(text):
    """
    :param text:
    :type text: str
    :return: True if text has any space or false otherwise.
    :rtype: bool
    """
    words = text.split(" ")
    if len(words) > 1:
        return True
    else:
        return False


def load_configuration():
    """ Read configuration file and populate with its data a
    config.Configuration instance.

    :return: Configuration instance populated with configuration file data.
    :rtype: config.Configuration
    """
    try:
        configuration = _read_config_file()
    except ConfigNotFound:
        _create_default_config_file()
        configuration = load_configuration()
    return configuration


def _read_config_file():
    """ Load all configuration parameters set in config file.

    :return: Configuration instance populated with configuration file data.
    :rtype: config.Configuration
    :raise: config.ConfigNotFound
    """
    try:
        with open(CONFIG_FILE_PATH, "rb") as config_file:
            configuration = pickle.load(config_file)
    except FileNotFoundError:
        raise ConfigNotFound()
    return configuration


def _create_default_config_file():
    """ Create a default configuration file.

    :return: None
    """
    default_configuration = Configuration()
    save_configuration(default_configuration)


def save_configuration(configuration):
    """ Write Configuration object in config file.

    :param configuration: Configuration to be saved.
    :type configuration: config.Configuration
    :return: None
    """
    with open(CONFIG_FILE_PATH, "wb") as config_file:
        pickle.dump(configuration, config_file, pickle.HIGHEST_PROTOCOL)


def _get_folder_path(path):
    """ If path is relative, get absolute path of current working directory
    suffixed by path. If path is absolute, just return it.

    :param path: Path to get absolute form.
    :type path: str
    :return: Absolute path.
    :rtype: str
    """
    absolute_directory = None
    if path.startswith("/"):
        absolute_directory = path
    else:
        current_working_directory = os.getcwd()
        absolute_directory = "{0}/{1}".format(current_working_directory,
                                              path)
    return absolute_directory


def _unknown_locators(locator_list):
    """ Detects if any locator in provided list is not registered as a valid one.

    Enabled locators are registered in DEFAULT_LOCATORS_PREFERENCE constant.
    Locators have to be one of them to be declared valid.

    :param locator_list: String list with locator names.
    :type locator_list: list
    :return: True if any locator in list is not within default locator list, else False.
    :rtype: bool
    """
    locator_set = set(locator_list)
    default_locator_set = set(DEFAULT_LOCATORS_PREFERENCE)
    if locator_set <= default_locator_set:
        return False
    else:
        return True


def _get_unknown_locators(locator_list):
    """
    :param locator_list: String list with locator names.
    :type locator_list: list
    :return: Set with unknown locators detected.
    :rtype: set
    """
    locator_set = set(locator_list)
    default_locator_set = set(DEFAULT_LOCATORS_PREFERENCE)
    return locator_set - default_locator_set


class ConfigNotFound(Exception):
    """ Launched when config file is not where is supposed to be."""
    def __init__(self):
        message = "Configuration file is not at it's default " \
                  "location: {0}".format(CONFIG_FILE_PATH)
        Exception.__init__(self, message)


class ParameterNotValid(Exception):
    """ Launched when user_id validation fails."""
    def __init__(self, provided_value, parameter, message):
        self.provided_value = provided_value
        self.parameter = parameter
        parameter_message = "There is a problem with parameter {0}, you " \
                            "gave {1} as value.\n " \
                            "Problem is: \n".format(parameter, provided_value)
        final_message = "".join([parameter_message, message])
        Exception.__init__(self, final_message)


class UnknownLocators(Exception):
    """ Raised when an still not implemented location is referenced in any
    operation.
    """

    def __init__(self, unknown_locators):
        unknown_locators_text = " ".join(unknown_locators)
        self.message = " ".join(["You tried to use not implemented locators:",
                                 unknown_locators_text])
        Exception.__init__(self, self.message)


class OpenConfigurationToUpdate(object):
    """ Context manager to get a configuration file and save it automatically.
    """
    def __init__(self):
        self.configuration = load_configuration()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        save_configuration(self.configuration)
        if exc_type is None:
            return True
        else:
            return False





"""
 geolocate wrapper to Geolite2 API.

 Programmed by: Dante Signal31

 email: dante.signal31@gmail.com
"""
import abc
import datetime
import gzip
import os
import shutil
# import subprocess
import tempfile
import geoip2.database as database
import geoip2.webservice as webservice
import maxminddb
import wget


import geolocate.classes.config as config
import geolocate.classes.exceptions as exceptions

DEFAULT_DATABASE_FILE_EXTENSION = "mmdb"
GEOIP2_WEBSERVICE_TAG = "geoip2_webservice"
GEOIP2_LOCAL_TAG = "geoip2_local"


def load_geoip_database(configuration=None):
    return GeoIPDatabase(configuration)


class GeoIPDatabase(object):
    """ Location engines may have multiple query methods. This class
    encapsulates them all in _locators list.
    """
    def __init__(self, configuration):
        """
        :param configuration: Geolocate configuration.
        :type configuration: config.Configuration
        """
        self._configuration = configuration
        self._locators = {}
        self._add_locators()
        self._locators_preference = configuration.locators_preference

    def _add_locators(self):
        """ Add query methods for this location engine.

        :return: None
        """
        self._add_webservice_locator()
        self._add_local_database_locator()

    def _web_service_access_configured(self):
        """
        :return: True if access credential to web service are configured, False if not.
        :rtype: bool
        """
        default_user_id = config.DEFAULT_USER_ID
        default_license_key = config.DEFAULT_LICENSE_KEY
        if self._configuration.user_id != default_user_id and \
           self._configuration.license_key != default_license_key:
            return True
        else:
            return False

    def _add_webservice_locator(self):
        """
        :return: None
        """
        if self._web_service_access_configured():
            webservice_locator = WebServiceGeoLocator(self._configuration)
            self._locators[GEOIP2_WEBSERVICE_TAG] = webservice_locator

    def _add_local_database_locator(self):
        """
        :return: None
        """
        local_db_locator = LocalDatabaseGeoLocator(self._configuration)
        self._locators[GEOIP2_LOCAL_TAG] = local_db_locator

    @property
    def geoip2_webservice(self):
        """
        :return: GeoIPLocateor to query GeoIP webservice.
        :rtype: WebServiceGeoLocator
        :raises: GeoIP2WebServiceNotConfigured
        """
        try:
            return self._locators[GEOIP2_WEBSERVICE_TAG]
        except KeyError:
            raise GeoIP2WebServiceNotConfigured()

    @property
    def geoip2_local(self):
        return self._locators[GEOIP2_LOCAL_TAG]

    def locate(self, ip):
        """ Query enabled locators in preference order until getting any
        geodata.

        :param ip: IP address to look for.
        :type ip: IP address string.
        :return: Location data for that address.
        :rtype: geoip2.models.City
        """
        for locator_id in self._locators_preference:
            # TODO: Try to disable webservice to see if this is really working.
            try:
                locator = self._locators[locator_id]
                geodata = locator.locate(ip)
            except:
                continue
            else:
                break
        else:
            raise exceptions.IPNotFound(ip)
        return geodata


class GeoLocator(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def __init__(self, configuration):
        self._configuration = configuration
        self._db_connection = None

    def locate(self, ip):
        """ Get geolocation data from database.

        :param ip: IP address we are asking about.
        :type ip: str
        :raises: geoip2.errors.AddressNotFoundError
        :return: Geolocation data.
        :rtype: geoip2.models.City
        """
        geolocation_data = self._db_connection.city(ip)
        return geolocation_data


class WebServiceGeoLocator(GeoLocator):
    def __init__(self, configuration):
        """
        :param configuration: Geolocate configuration.
        :type configuration: config.Configuration
        :return: None
        """
        super().__init__(configuration)
        self._db_connection = webservice.Client(configuration.user_id,
                                                configuration.license_key)


class LocalDatabaseGeoLocator(GeoLocator):
    def __init__(self, configuration):
        """
        :param configuration: Geolocate configuration.
        :type configuration: config.Configuration
        :return: none
        :raise: LocalDatabaseNotFound
        :raise: InvalidLocalDatabase
        """
        super().__init__(configuration)
        self._update_db()
        db_path = configuration.local_database_path
        self._db_connection = _open_local_database(db_path)

    def _update_db(self):
        """ Download a fresh geolocation database if current is too old.

        :return: None
        """
        if self._local_database_too_old():
            self._download_fresh_database()

    def _local_database_too_old(self):
        """
        :return: True if database file has to be refreshed, False if not.
        :rtype: bool
        """
        database_path = self._configuration.local_database_path
        update_interval = self._configuration.update_interval
        try:
            last_modification = _get_database_last_modification(database_path)
        except LocalDatabaseNotFound:
            return True  # This should force a database download.
        else:
            today_date = datetime.date.today()
            allowed_age = datetime.timedelta(days=update_interval)
            return _must_be_updated(today_date, last_modification, allowed_age)

    def _download_fresh_database(self):
        """ Download compressed database, decompress it and place it instead
        old one.

        :return: None
        """
        with tempfile.TemporaryDirectory() as temporary_directory:
            print("Downloading fresh geolocation database...")
            self._download_file(temporary_directory)
            try:
                _decompress_file(temporary_directory)
            except CompressedFileNotFound as e:
                _print_compressed_file_not_found_error(e)
            else:
                self._write_new_database(temporary_directory)

    def _download_file(self, temporal_directory):
        """
        :param temporal_directory: Folder path to place downloaded file in.
        :type temporal_directory: str
        :return: None
        """
        wget.download(url=self._configuration.download_url,
                      out=temporal_directory)

    def _remove_old_database(self):
        """
        :return: None
        """
        database_path = self._configuration.local_database_path
        os.remove(database_path)

    def _write_new_database(self, temporary_directory):
        """
        :param temporary_directory: Folder path to place downloaded file in.
        :type temporal_directory: str
        :return: None
        """
        try:
            self._remove_old_database()
        except FileNotFoundError:
            print("\nOld local database not found. May be this is the "
                  "first time you run geolocate?.")
        self._copy_new_database(temporary_directory)

    def _copy_new_database(self, decompressed_file_path):
        """
        :param decompressed_file_path: Folder where new database is placed.
        :type decompressed_file_path: str
        :return: None
        """
        database_path = self._configuration.local_database_path
        new_database_path = _get_new_database_path_name(decompressed_file_path)
        shutil.copyfile(new_database_path, database_path)


def _decompress_file(temporary_directory):
    """ Decompress tar.gz file found in temporary_directory.

    :param temporary_directory: Folder path to compressed file.
    :type temporary_directory: str
    :return: Path to decompressed folder.
    :rtype: str
    """
    try:
        compressed_file_name_path = _find_compressed_file(temporary_directory)
        uncompressed_file_name_path = _get_uncompressed_file_name_path(compressed_file_name_path)
    except CompressedFileNotFound as e:
        _print_compressed_file_not_found_error(e)
    else:
        with gzip.open(compressed_file_name_path, "rb") as input_file, \
                open(uncompressed_file_name_path, "wb") as output_file:
            uncompressed_content = input_file.read()
            output_file.write(uncompressed_content)
        return compressed_file_name_path


def _find_compressed_file(temporary_directory):
    """ Find .gz file name downloaded to temporary directory.

    :param temporary_directory: Folder to search compressed file in.
    :type temporary_directory: str
    :return: Absolute path name of found file.
    :rtype: str
    :raise: CompressedFileNotFound
    """
    for file_name in os.listdir(temporary_directory):
        if file_name.endswith(".gz"):
            file_name_path = os.path.join(temporary_directory, file_name)
            return file_name_path
    else:
        raise CompressedFileNotFound(temporary_directory)

def _get_uncompressed_file_name_path(compressed_file_name_path):
    """ Get file name compressed in .gz file.

    :param compressed_file_name_path: Compressed file name absolute path.
    :type compressed_file_name_path: str
    :return: File name compressed into .gz file.
    :rtype: str
    """
    uncompressed_file_name_path = os.path.splitext(compressed_file_name_path)[0]
    return uncompressed_file_name_path

def _open_local_database(local_database_path):
    try:
        database_connection = database.Reader(local_database_path)
    except FileNotFoundError:
        raise LocalDatabaseNotFound(local_database_path)
    except maxminddb.InvalidDatabaseError:
        raise InvalidLocalDatabase(local_database_path)
    else:
        return database_connection


def _get_database_last_modification(database_path):
    """
    :param database_path: Path to database file to be evaluated.
    :type database_path: str
    :return: Date of file's last modification.
    :rtype: datetime.date
    """
    try:
        last_modification = os.stat(database_path).st_mtime
    except FileNotFoundError:
        raise LocalDatabaseNotFound(database_path)
    else:
        date_last_modification = datetime.date.fromtimestamp(last_modification)
        return date_last_modification


def _must_be_updated(today_date, last_modification, allowed_age):
    """
    :param today_date: Today's date.
    :type today_date: datetime.date
    :param last_modification: Date of file's last modification.
    :type last_modification: datetime.date
    :param allowed_age: Maximum age allowed between today and last modification.
    :type allowed_age: datetime.timedelta
    :return: True if last modification is older than allowed age; else False.
    :rtype: bool
    """
    if today_date - last_modification > allowed_age:
        return True
    else:
        return False


def _get_new_database_path_name(decompressed_file_path):
    """
    :param decompressed_file_path: Temporary folder path where new database is.
    :type decompressed_file_path: str
    :return: Temporary folder with database filename and extension appended.
    :rtype: str
    """
    decompressed_files = os.listdir(decompressed_file_path)
    for file in decompressed_files:
        if file.endswith(DEFAULT_DATABASE_FILE_EXTENSION):
            new_database_path_name = os.path.join(decompressed_file_path, file)
            return new_database_path_name
    else:
        raise NotValidDatabaseFileFound(decompressed_file_path)


def _print_compressed_file_not_found_error(e):
    """
    :param e: Exception caught.
    :type e: CompressedFileNotFound
    :return: none
    """
    print("Problem decompressing updated database.")
    path = e.compressed_database_path
    message = "No .gz file found at {0}".format(path)
    print(message)


class GeoIP2WebServiceNotConfigured(Exception):
    """ GeoIP2 WebService access is still not configured."""

    def __init__(self):
        message = "You tried a query to GeoIP2 webservice, but no valid " \
                  "credentials were found in configuration."
        Exception.__init__(self, message)


class LocalDatabaseNotFound(OSError):
    """ Local database file is missing."""

    def __init__(self, local_database_path):
        self.local_database_path = local_database_path
        message = "Local database file is missing"
        OSError.__init__(self, message)


class InvalidLocalDatabase(Exception):
    """ Local database exists but is corrupted. """

    def __init__(self, local_database_path):
        self.local_database_path = local_database_path
        message = "Local database is invalid."
        Exception.__init__(self, message)


class NotValidDatabaseFileFound(OSError):
    """ Raised when a new database pack is downloaded on local, but after
    decompression no valid database file is found in decompressed folder.
    """

    def __init__(self, decompressed_database_path):
        self.decompressed_database_path = decompressed_database_path
        message = "No valid database found in downloaded file."
        OSError.__init__(self, message)


class CompressedFileNotFound(OSError):
    """ Raised when no .gz compressed file is found in temporary folder where
    downloaded data is placed.
    """

    def __init__(self, compressed_database_path):
        self.compressed_database_path = compressed_database_path
        message = "No compressed file found in downloaded data."
        OSError.__init__(self, message)


"""
 Custom exceptions for geolocate.

 Programmed by: Dante Signal31

 email: dante.signal31@gmail.com
"""

# TODO: Find the way to take this code back to geowrapper module.
# This class shouldn't be here. Initially it was defined at geowrapper module.
# Problem is that when class is there, test_GeolocateInputParser_next (at
# test_parser.TestParser) fails because IPNotFound is not caught in parser's
# GeoLocateInputParser __next__(). I don't know why but if I take off IPNotFound
# definition from geowrapper module and put it in another module, like this one,
# the code runs as intended and tests success.
class IPNotFound(Exception):
    """ Searched IP is not in geolocation database."""

    def __init__(self, ip_address):
        self.failed_IP = ip_address
        self.message = "The address {0} is not in the database.".format(ip_address)
        Exception.__init__(self, self.message)


"""
 geolocate parser

 Programmed by: Dante Signal31

 email: dante.signal31@gmail.com
"""

import re
import sys
import collections

import geolocate.classes.exceptions as exceptions


class GeolocateInputParser(object):

    _VERBOSITY_FIELDS = ["continent_name",
                         "country_name",
                         "city_name",
                         "lat-long"]
    VERBOSITY_LEVELS = range(len(_VERBOSITY_FIELDS))
    _IP_NOT_FOUND_MESSAGE = "[IP not found]"

    def __init__(self, verbosity, geoip_database, text=None):
        self._verbosity = verbosity
        self._geoip_database = geoip_database
        if text is None:
            self._entered_text = InputReader()
        else:
            self._entered_text = _get_lines(text)

    def __iter__(self):
        return self

    def __next__(self):
        """Read a line from entered text, search for an IP and get
        their geolocation data.
        """
        # We could analyze the whole text but it is expected that input
        # will come from another program's output through a shell pipe,
        # so in that case is more efficient reading lines as they
        # arrive from the pipe.
        for line in self._entered_text:
            try:
                ip_list = _find_ips_in_text(line)
                for ip in ip_list:
                    location = self._locate(ip)
                    line = _include_location_in_line(line, ip, location)
                return line
            except exceptions.IPNotFound as e:
                location_string = _join_ip_to_location(e.failed_IP,
                                                       self.__class__._IP_NOT_FOUND_MESSAGE)
                line = _include_location_in_line(line, e.failed_IP,
                                                 location_string)
                return line
        raise StopIteration()

    def _locate(self, ip):
        """Query database to get IP address location and format
        location depending of desired verbosity.

        :param ip: String with IP address.
        :type ip: str
        :return: String with location.
        :rtype: str
        :raises: exceptions.IPNotFound.
        """
        try:
            location_data = self._geoip_database.locate(ip)
        except exceptions.IPNotFound:
            raise
        location_string = self._format_location_string(location_data)
        returned_string = _join_ip_to_location(ip, location_string)
        return returned_string

    def _format_location_string(self, location_data):
        """Add location fields to returned string depending of desired
        verbosity.

        :param location_data: GeoIP record.
        :type location_data: geoip2.models.City
        :return: String with location.
        :rtype: str
        """
        # TODO: This code sucks, I have to refactor it.
        location_string = "["
        verbosity_levels_to_show = range(self._verbosity+1)
        VERBOSITY_FIELDS = self.__class__._VERBOSITY_FIELDS
        location_fields = _find_unknowns(location_data)
        # The more verbosity the more data you append to returned string.
        for i in verbosity_levels_to_show:
            if VERBOSITY_FIELDS[i] == "lat-long":
                lat_long = ", ".join([location_fields["lat-long"]["latitude"],
                                      location_fields["lat-long"]["longitude"]])
                location_string = " | ".join([location_string,
                                              lat_long])
            elif VERBOSITY_FIELDS[i] == "continent_name":
                # Continent_name is first level of verbosity so it is not
                # prepended by an "|"
                location_string = "".join([location_string,
                                          location_fields["continent_name"]])
            elif VERBOSITY_FIELDS[i] == "city_name":
                location_string = " | ".join([location_string,
                                              location_fields["city_name"]])
            elif VERBOSITY_FIELDS[i] == "country_name":
                location_string = " | ".join([location_string,
                                              location_fields["country_name"]])
        location_string = "".join([location_string, "]"])
        return location_string


def _find_unknowns(location_data):
    """ Find any unknown attribute and convert it in a informational string.

    :param location_data: GeoIP record.
    :type location_data: geoip2.models.City
    :return: Informational strings.
    :rtype: dict
    """
    location_strings = _default_location_strings()
    if location_data.continent.name is None:
        location_strings["continent_name"] = "Unknown continent"
    else:
        location_strings["continent_name"] = location_data.continent.name
    if location_data.country.name is None:
        location_strings["country_name"] = "Unknown country"
    else:
        location_strings["country_name"] = location_data.country.name
    if location_data.city.name is None:
        location_strings["city_name"] = "Unknown city"
    else:
        location_strings["city_name"] = location_data.city.name
    if location_data.location.latitude is None:
        location_strings["lat-long"]["latitude"] = "Unknown latitude"
    else:
        location_strings["lat-long"]["latitude"] = str(location_data.location.latitude)
    if location_data.location.longitude is None:
        location_strings["lat-long"]["longitude"] = "Unknown longitude"
    else:
        location_strings["lat-long"]["longitude"] = str(location_data.location.longitude)
    return location_strings


def _default_location_strings():
    """ Generates a dict object suitable for location strings.

    :return: location strings object.
    :rtype: dict
    """
    location_strings = collections.defaultdict()
    location_strings["lat-long"] = collections.defaultdict()
    return location_strings


def _find_ips_in_text(text):
    """Return a set with all IP addresses found in text

    :param text: Text with IP addresses embedded.
    :type text: str
    :return: A set with all addresses found.
    :rtype: set
    """
    ipv4_addresses = _find_ipv4_addresses(text)
    # TODO: Implement a _find_ipv6_addresses() so in the end we can do:
    #          found_addresses = ipv4_addresses.union(ipv6_addresses)
    found_addresses = ipv4_addresses
    return found_addresses


def _find_ipv4_addresses(text):
    """
    :param text: Text with IP addresses embedded.
    :type text: str
    :return: A set with all addresses found.
    :rtype: set
    """
    ip_filter = "[0-9]+(?:\.[0-9]+){3}"
    regex_filter = re.compile(ip_filter)
    addresses = regex_filter.findall(text)
    return set(addresses)


def _include_location_in_line(line, ip, location):
    """
    :param line: Original line to place location string into.
    :type line: str
    :param ip: IP address to be replaced with location.
    :type ip: str
    :param location: IP address string with location data appended.
    :type location: str
    :return: Line with ip addresses followed by location strings.
    :rtype: str
    """
    return line.replace(ip, location)


def _join_ip_to_location(ip, location):
    """
    :param ip: IP address.
    :type ip: str
    :param location: Location data string.
    :return: IP address with location data string appended.
    :rtype: str
    """
    location_string = " ".join([ip, location])
    return location_string


def _get_lines(text):
    """ Get a generator object with text lines.

    :param text: Text to parse.
    :type text: str
    :return: Text lines.
    :rtype: generator
    """
    lines = (line for line in text.split("\n"))
    return lines


class InputReader(object):
    """Iterator to read piped input from other programs."""

    def __init__(self):
        pass

    def __iter__(self):
        return self

    @staticmethod
    def __next__():
        for line in sys.stdin:
            return line
        raise StopIteration()


#!/usr/bin/env python3
"""
 geolocate

 Programmed by: Dante Signal31

 email: dante.signal31@gmail.com

This scripts scan given text to find urls and IP addresses. The output is the
same text but every url and IP address is going to have its geolocation
appended.

Geolocate is possible thanks to `Maxmind GeoIP database <http://www.maxmind.com>`_
and their API.
"""

import geolocate.classes.arguments as arguments
import geolocate.classes.geowrapper as geowrapper
import geolocate.classes.parser as parser
import geolocate.classes.config as config

import sys

def print_lines_parsed(parser):
    for line in parser:
        print(line, end="")


def main():
    _arguments = arguments.parse_arguments()
    arguments.process_optional_parameters(_arguments)
    configuration = config.load_configuration()
    geoip_database = geowrapper.load_geoip_database(configuration)
    if _arguments.text_to_parse or _arguments.stream_mode:
        input_parser = parser.GeolocateInputParser(_arguments.verbosity,
                                                   geoip_database,
                                                   _arguments.text_to_parse)
        print_lines_parsed(input_parser)
    print()

if __name__ == "__main__":
    main()


from setuptools import setup, find_packages  # Always prefer setuptools over distutils

long_description = """This program accepts any text and searchs inside every IP address. With
each of those IP addresses, geolocate queries Maxmind GeoIP database (http://www.maxmind.com)_
to look for the city and country where IP address or URL is located.

Geolocate is designed to be used in console with pipes and redirections along
with applications like traceroute, etc. Geolocate's output is the same text
than input but IP addresses are going to have appended its country and city and long-lat
coordinates (depending on verbosity level).

More info in: https://bitbucket.org/dante_signal31/geolocate
"""

setup(name="glocate",
      version="1.2.0",
      description="This program scans given text to find urls and IP addresses. "
                  "The output is the same text but every url and IP address "
                  "is going to have its geolocation data appended.",
      long_description=long_description,
      author="Dante Signal31",
      author_email="dante.signal31@gmail.com",
      license="BSD-3",
      url="https://github.com/dante-signal31/geolocate",
      download_url="https://github.com/dante-signal31/geolocate/releases",
      classifiers=['Development Status :: 4 - Beta',
                   'Intended Audience :: Developers',
                   'Intended Audience :: Information Technology',
                   'Intended Audience :: Science/Research',
                   'Intended Audience :: System Administrators',
                   'Intended Audience :: Telecommunications Industry',
                   'Intended Audience :: Other Audience',
                   'Topic :: System :: Networking',
                   'Topic :: Security',
                   'License :: OSI Approved :: BSD License',
                   'Operating System :: POSIX :: Linux',
                   'Programming Language :: Python :: 3',
                   'Programming Language :: Python :: 3.4'],
      keywords="geolocation ip addresses",
      install_requires=["geoip2>=2.1.0", "maxminddb>=1.1.1", "requests>=2.5.0",
                        "wget>=2.2"],
      zip_safe=False,
      # TODO: This exclude is not working when building wheels, tests package
      # is still included in packages. It's a bug in pip:
      #     https://bitbucket.org/pypa/wheel/issue/99/cannot-exclude-directory
      # Until it is fixed, the workaround is to compile wheel package in two
      # steps:
      #     python setup.py sdist
      #     pip wheel --no-index --no-deps --wheel-dir dist dist/*.tar.gz
      packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*",
                                      "tests", "*tests*"]),
      entry_points={'console_scripts': ['geolocate=geolocate.glocate:main', ],
                    },
      package_data={"geolocate": ["etc/empty.txt",
                                  "local_database/empty.txt"], }
      )



"""
 Geolocate argument parser module.

 Programmed by: Dante Signal31

 email: dante.signal31@gmail.com
"""
import argparse
import sys

import geolocate.classes.config as config
import geolocate.classes.parser as parser


# These arguments activate functions with their same name. True or False depends
# on passing in user arguments to called function or not.
MUST_PASS_IN_USER_ARGUMENTS = {"show_enabled_locators": False,
                                "set_locators_preference": True,
                                "show_disabled_locators": False,
                                "reset_locators_preference": False,
                                "set_user": True,
                                "set_password": True,
                                "show_user": False,
                                "show_password": False}
# This arguments don't activate a function with their same name.
NOT_CALLABLE_ARGUMENTS = {"verbosity", "text_to_parse", "stream_mode"}


def parse_arguments():
    verbosity_choices = parser.GeolocateInputParser.VERBOSITY_LEVELS
    arg_parser = argparse.ArgumentParser(description="Locate IP adresses "
                                                     "in given text.\n",
                                         epilog="This program is possible "
                                                "thanks to Maxmind GeoIP "
                                                "database "
                                                "<http://www.maxmind.com> and "
                                                "their API.")
    data_input_arguments = arg_parser.add_mutually_exclusive_group()
    data_input_arguments.add_argument(dest="text_to_parse",
                                      metavar="\"text to parse\"",
                                      nargs="?", type=str, default=None,
                                      help="Text to analyze surrounded by "
                                           "double quotes.")
    data_input_arguments.add_argument("-s", "--stream",
                                      dest="stream_mode",
                                      action="store_true", default=False,
                                      help="Program will analyze piped output "
                                           "from another program.")
    arg_parser.add_argument("-v", "--verbosity", dest="verbosity",
                            choices=verbosity_choices, type=int, default=0,
                            help="0-3 The higher the more geodata.")
    arg_parser.add_argument("-l", "--show_enabled",
                            dest="show_enabled_locators",
                            action="store_true", default=False,
                            help="Show enabled locators ordered by preference.")
    arg_parser.add_argument("-p", "--set_preference",
                            dest="set_locators_preference",
                            nargs="*", default=None,
                            help="Set preferred locator order.",
                            metavar="locator")
    arg_parser.add_argument("-d", "--show_disabled",
                            dest="show_disabled_locators", action="store_true",
                            default=False, help="Show disabled locators.")
    arg_parser.add_argument("-r", "--reset", dest="reset_locators_preference",
                            action="store_true", default=False,
                            help="Restore default locator order.")
    arg_parser.add_argument("-u", "--set_user",
                            dest="set_user",
                            type=str, default=None,
                            help="Set user for webservice database access.",
                            metavar="account_id")
    arg_parser.add_argument("-w", "--set_password",
                            dest="set_password",
                            type=str, default=None,
                            help="Set password for webservice database access.",
                            metavar="account_password")
    arg_parser.add_argument("-i", "--show_user",
                            dest="show_user", action="store_true",
                            default=False, help="Show user configured for "
                                                "webservice access.")
    arg_parser.add_argument("-k", "--show_password",
                            dest="show_password", action="store_true",
                            default=False, help="Show password configured for "
                                                "webservice access.")
    return arg_parser.parse_args()


def show_enabled_locators():
    """ Print in console enabled locators ordered by preference.

    :return: None
    """
    enabled_locators = _get_enabled_locators_list()
    _print_locators_list("Enabled locators:", enabled_locators)


def set_locators_preference(arguments):
    """ Change locators list with new list given in arguments.

    :param arguments:  Arguments object returned by ArgumentParser.parse_args()
    :type arguments: Namespace
    :return: None
    """
    with config.OpenConfigurationToUpdate() as f:
        f.configuration.locators_preference = arguments.set_locators_preference


def show_disabled_locators():
    """ Print in console enabled locators ordered by preference.

    :return: None
    """
    enabled_locators_set = set(_get_enabled_locators_list())
    complete_locators_set = set(config.DEFAULT_LOCATORS_PREFERENCE)
    disabled_locators_set = complete_locators_set - enabled_locators_set
    _print_locators_list("Disabled locators:", disabled_locators_set)


def reset_locators_preference():
    """ Sets back locators list to it's default value.

    :return: None
    """
    with config.OpenConfigurationToUpdate() as f:
        f.configuration.reset_locators_preference()


def _get_enabled_locators_list():
    """
    :return: Configuration locators preference.
    :rtype: list
    """
    configuration = config.load_configuration()
    enabled_locators = configuration.locators_preference
    return enabled_locators


def set_user(arguments):
    """ Set user for webservice database access.

    :param arguments:  Arguments object returned by ArgumentParser.parse_args()
    :type arguments: Namespace
    :return: None
    """
    with config.OpenConfigurationToUpdate() as f:
        f.configuration.user_id = arguments.set_user


def set_password(arguments):
    """ Set password for webservice database access.

    :param arguments:  Arguments object returned by ArgumentParser.parse_args()
    :type arguments: Namespace
    :return: None
    """
    with config.OpenConfigurationToUpdate() as f:
        f.configuration.license_key = arguments.set_password


def show_user():
    """ Print on console user configured for webservice database access.

    :return: None
    """
    configuration = config.load_configuration()
    output_string = "User:\n" \
                    "{0}".format(configuration.user_id)
    print(output_string)


def show_password():
    """ Print on console password configured for webservice database access.

    :return: None
    """
    configuration = config.load_configuration()
    output_string = "Password:\n" \
                    "{0}".format(configuration.license_key)
    print(output_string)



def _print_locators_list(header, locators_list):
    """ Print on console a formatted list of locators.

    :param header: Title of list.
    :type header: str
    :param locators_list: Locators list.
    :type locators_list: list or set
    :return: None
    """
    components = list(locators_list)
    components.insert(0, header)
    message = "\n".join(components)
    print(message)


def process_optional_parameters(arguments):
    """ Take all optional arguments given by user and run one by one all
    functions assigned to each of them.

    :param arguments: Arguments object returned by ArgumentParser.parse_args()
    :type arguments: Namespace
    :return: None
    """
    functions_to_execute = _get_functions_to_execute(arguments)
    for functions in functions_to_execute:
        try:
            _execute_function(functions, arguments)
        except (NoFunctionAssignedToArgument, config.UnknownLocators) as e:
            sys.exit(". ".join([e.message, "Exiting"]))


def _get_functions_to_execute(arguments):
    """
    :param arguments: Arguments object returned by ArgumentParser.parse_args()
    :type arguments: Namespace
    :return: Function names set to execute.
    :rtype: set
    """
    valid_arguments = _get_user_arguments(arguments)
    functions_names_set = {argument for argument in valid_arguments
                           if getattr(arguments, argument)}
    return functions_names_set


def _get_user_arguments(arguments):
    """ Get public user attributes of arguments object.

    Arguments set by user are defined public in object returned by
    ArgumentParser.parse_args(). Built-in attributes in that object are defined
    "private" with "_" o "__" prefix.

    :param arguments: Arguments object returned by ArgumentParser.parse_args()
    :type arguments: Namespace
    :return: Public attribute list of arguments object.
    :rtype: set
    """
    attribute_list = dir(arguments)
    public_attributes_set = {argument for argument in attribute_list
                             if not argument.startswith("_")}
    public_attributes_set = _remove_non_user_attributes(public_attributes_set)
    return public_attributes_set


def _remove_non_user_attributes(attributes_set):
    """ Takes out from attribute set public attributes not defined by user.

    :param attributes_set: Public attribute list of arguments object.
    :type attributes_set: set
    :return: Public attribute list of arguments object without non user attributes.
    :rtype: set
    """
    # Arguments object returned by ArgumentParser.parse_args() include by
    # default "index" and "count" public attributes. We don't need them so we
    # strip them off. Arguments "verbosity", "text_to_parse" and "stream_mode"
    # is not an optional parameter and shouldn't be processed as one.
    non_user_attributes = {"index", "count"} | NOT_CALLABLE_ARGUMENTS
    user_attributes = attributes_set.difference(non_user_attributes)
    return user_attributes


def _execute_function(argument, arguments):
    """ Execute function assigned to given argument.

    :param argument: Argument to search a function assigned for.
    :param arguments: Arguments object returned by ArgumentParser.parse_args()
    :type arguments: Namespace
    :return: None
    :raise: NoFunctionAssignedToArgument
    """
    try:
        if MUST_PASS_IN_USER_ARGUMENTS[argument]:
            # Arguments names are equal to names of functions to call.
            _call_function(argument, arguments)
        else:
            _call_function(argument)
    except KeyError:
        raise NoFunctionAssignedToArgument(argument)


def _call_function(function_name, arguments=None):
    """ Call given function and pass it in arguments if they are given.

    :param function_name: Function to call.
    :type function_name: str
    :param arguments:  Arguments object returned by ArgumentParser.parse_args()
    :type arguments: Namespace
    :return: None
    """
    function = globals().get(function_name)
    if arguments is None:
        function()
    else:
        function(arguments)


class NoFunctionAssignedToArgument(Exception):
    """ User has given and argument that has no function assigned yet."""

    def __init__(self, argument):
        self.argument = argument
        self.message = "Sorry, I don't know what to do " \
                       "with argument: {0}".format(self.argument)
        Exception.__init__(self, self.message)